from agentine.llm import LLM
from agentine.message import Message, FileMessage
from agentine.agent import Agent, AgentGroup, AgentIndex, AgentLegion
from agentine.chatbot import Chatbot
from .config import (
    DEFAULT_LLM_PROVIDER,
    KNOWN_LLM_PROVIDERS,
)

__all__ = [
    "LLM",
    "Message",
    "FileMessage",
    "Agent",
    "AgentGroup",
    "AgentIndex",
    "AgentLegion",
    "Chatbot",
    "DEFAULT_LLM_PROVIDER",
    "KNOWN_LLM_PROVIDERS",
]


from agentine import Message, LLM

msg = Message(
    role="user",
    content=[
        {
            "type": "text",
            "text": "Describe what is in this image:",
        },
        {
            "type": "image_url",
            "image_url": {
                "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/"
                "Gfp-wisconsin-madison-the-nature-boardwalk.jpg/"
                "2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg"
            },
        },
        {
            "type": "text",
            "text": "Then list the color found in the image.",
        },
    ],
)

llm = LLM(provider="openai")
result = llm.chat(messages=[msg])

print(result)


from collections.abc import Iterator, AsyncIterator
from pydantic import BaseModel, Field, PrivateAttr
from openai import OpenAI, AsyncOpenAI
import asyncio
from concurrent.futures import ThreadPoolExecutor

from agentine.config import DEFAULT_LLM_PROVIDER, KNOWN_LLM_PROVIDERS
from agentine.message import Message


class LLM(BaseModel):
    api_key: str | None = Field(default=None)
    provider: str = Field(default=DEFAULT_LLM_PROVIDER)
    model: str | None = None
    base_url: str = ""

    timeout: int = 60000
    max_retries: int = 2
    max_concurrency: int = 100

    response_format: str = Field(default="text")
    temperature: float | None = None
    max_completion_tokens: int | None = None
    reasoning_effort: str | None = None

    _client: OpenAI = PrivateAttr()
    _client_async: AsyncOpenAI = PrivateAttr()

    class Config:
        extra = "allow"

    def __init__(self, **data: object):
        if "provider" not in data:
            data["provider"] = DEFAULT_LLM_PROVIDER

        super().__init__(**data)
        self._apply_provider_config(self.provider)
        self._init_clients()

    def __setattr__(self, name: str, value: object) -> None:
        super().__setattr__(name, value)

        if name == "provider":
            self._apply_provider_config(value)

        if name in {"provider", "api_key", "base_url"}:
            self._update_clients()

    def _apply_provider_config(self, provider: str) -> None:
        provider_key = str(provider).lower()
        provider_config = KNOWN_LLM_PROVIDERS.get(provider_key)

        if not provider_config:
            raise ValueError(
                f"422 - UNKNOWN PROVIDER: '{provider.upper()}'\n\n"
                f"Valid providers are: {', '.join(KNOWN_LLM_PROVIDERS.keys())}\n"
                f"If you want to use a custom provider, please set 'api_key' and 'base_url' manually instead."
            )

        super().__setattr__("model", provider_config.get("default_model"))
        super().__setattr__("api_key", provider_config.get("api_key"))
        super().__setattr__("base_url", provider_config.get("base_url"))

    def _init_clients(self) -> None:
        self._client = OpenAI(api_key=self.api_key, base_url=self.base_url)
        self._client_async = AsyncOpenAI(api_key=self.api_key, base_url=self.base_url)

    def _update_clients(self) -> None:
        self._client.api_key = self.api_key
        self._client.base_url = self.base_url
        self._client_async.api_key = self.api_key
        self._client_async.base_url = self.base_url

    @property
    def client(self) -> OpenAI:
        return self._client

    @property
    def client_async(self) -> AsyncOpenAI:
        return self._client_async

    def completion_config(self) -> dict[str, object]:
        excluded = {
            "api_key",
            "provider",
            "base_url",
            "timeout",
            "max_retries",
            "max_concurrency",
        }
        data = self.model_dump(exclude=excluded, exclude_none=True)
        if "response_format" in data:
            data["response_format"] = {"type": data["response_format"]}
        return data

    def _prepare_kwargs(self, messages: list[Message]) -> dict[str, object]:
        assert isinstance(messages, list) and all(isinstance(m, Message) for m in messages)
        kwargs = self.completion_config()
        kwargs["model"] = self.model
        kwargs["messages"] = [m.core() for m in messages]
        return kwargs

    def chat(self, messages: list[Message]) -> Message:
        kwargs = self._prepare_kwargs(messages)
        for attempt in range(self.max_retries + 1):
            try:
                response = self.client.chat.completions.create(**kwargs)
                return Message.from_openai_completion(response)
            except Exception:
                if attempt == self.max_retries:
                    raise

    async def chat_async(self, messages: list[Message]) -> Message:
        kwargs = self._prepare_kwargs(messages)
        for attempt in range(self.max_retries + 1):
            try:
                response = await self.client_async.chat.completions.create(**kwargs)
                return Message.from_openai_completion(response)
            except Exception:
                if attempt == self.max_retries:
                    raise

    def stream(
        self, messages: list[Message], include_usage: bool = True
    ) -> Iterator[Message]:
        kwargs = self._prepare_kwargs(messages)
        kwargs["stream"] = True
        if include_usage:
            kwargs["stream_options"] = {"include_usage": True}
        stream = self.client.chat.completions.create(**kwargs)
        for chunk in stream:
            yield Message.from_openai_completion_chunk(chunk)

    async def stream_async(
        self, messages: list[Message], include_usage: bool = True
    ) -> AsyncIterator[Message]:
        kwargs = self._prepare_kwargs(messages)
        kwargs["stream"] = True
        if include_usage:
            kwargs["stream_options"] = {"include_usage": True}
        stream = await self.client_async.chat.completions.create(**kwargs)
        async for chunk in stream:
            yield Message.from_openai_completion_chunk(chunk)

    def batch(self, batch_messages: list[list[Message]]) -> list[Message | Exception]:
        def process(messages: list[Message]) -> Message | Exception:
            try:
                return self.chat(messages)
            except Exception as e:
                return e
        with ThreadPoolExecutor(max_workers=self.max_concurrency) as executor:
            return list(executor.map(process, batch_messages))

    async def batch_async(self, batch_messages: list[list[Message]]) -> list[Message | Exception]:
        async def process(messages: list[Message]) -> Message | Exception:
            try:
                return await self.chat_async(messages)
            except Exception as e:
                return e
        return await asyncio.gather(*(process(m) for m in batch_messages), return_exceptions=True)


import json
from collections.abc import AsyncIterator, Iterator
from typing import Literal
from pydantic import BaseModel, Field
import asyncio
from agentine.llm import LLM, Message
from agentine.utils import Utility


class Agent(BaseModel):
    instruction: str
    name: str | None = None
    llm: LLM = Field(default_factory=LLM)
    response_format: Literal["text", "json_schema", "json_object"] = Field(default="text")

    class Config:
        extra = "allow"

    def __init__(self, **data: object) -> None:
        super().__init__(**data)
        self.llm.response_format = self.response_format

    def __setattr__(self, key: str, value: object) -> None:
        super().__setattr__(key, value)
        if key == "response_format" and hasattr(self, "llm") and self.llm is not None:
            self.llm.response_format = value

    def _prepare_messages(
        self,
        query: str | None = None,
        messages: list[Message] | None = None,
        data: dict[str, object] | None = None,
    ) -> list[Message]:
        formatted_instruction = Utility.format(string=self.instruction, data=data) if data else self.instruction
        formatted_query = Utility.format(string=query, data=data) if data else query

        chat_messages: list[Message] = []

        if formatted_instruction:
            chat_messages.append(Message(role="system", content=formatted_instruction))
        if messages:
            chat_messages.extend(messages)
        if formatted_query:
            chat_messages.append(Message(role="user", content=formatted_query))

        return chat_messages

    def work(
        self,
        query: str | None = None,
        messages: list[Message] | None = None,
        data: dict[str, object] | None = None,
    ) -> Message:
        assert query is None or isinstance(query, str), "query must be a string or None"
        assert messages is None or isinstance(messages, list), "messages must be a list or None"
        assert data is None or isinstance(data, dict), "data must be a dict or None"

        chat_messages = self._prepare_messages(query=query, messages=messages, data=data)
        result = self.llm.chat(messages=chat_messages)

        if self.response_format == "json_object":
            result.data = json.loads(result.content)

        return result

    async def work_async(
        self,
        query: str | None = None,
        messages: list[Message] | None = None,
        data: dict[str, object] | None = None,
    ) -> Message:
        assert query is None or isinstance(query, str), "query must be a string or None"
        assert messages is None or isinstance(messages, list), "messages must be a list or None"
        assert data is None or isinstance(data, dict), "data must be a dict or None"

        chat_messages = self._prepare_messages(query=query, messages=messages, data=data)
        result = await self.llm.chat_async(messages=chat_messages)

        if self.response_format == "json_object":
            result.data = json.loads(result.content)

        return result

    def stream(
        self,
        query: str | None = None,
        messages: list[Message] | None = None,
        data: dict[str, object] | None = None,
    ) -> Iterator[Message]:
        assert query is None or isinstance(query, str), "query must be a string or None"
        assert messages is None or isinstance(messages, list), "messages must be a list or None"
        assert data is None or isinstance(data, dict), "data must be a dict or None"

        chat_messages = self._prepare_messages(query=query, messages=messages, data=data)
        return self.llm.stream(chat_messages)

    async def stream_async(
        self,
        query: str | None = None,
        messages: list[Message] | None = None,
        data: dict[str, object] | None = None,
    ) -> AsyncIterator[Message]:
        assert query is None or isinstance(query, str), "query must be a string or None"
        assert messages is None or isinstance(messages, list), "messages must be a list or None"
        assert data is None or isinstance(data, dict), "data must be a dict or None"

        chat_messages = self._prepare_messages(query=query, messages=messages, data=data)
        async for message in self.llm.stream_async(chat_messages):
            yield message


class AgentGroup(BaseModel):
    agents: list[Agent] = Field(default_factory=list)

    def work(
        self,
        query: str,
        messages: list[Message] | None = None,
        data: dict[str, object] | None = None,
    ) -> list[Message]:
        results = []
        for agent in self.agents:
            result = agent.work(query=query, messages=messages, data=data)
            results.append(result)
        return results

    async def work_async(
        self,
        query: str,
        messages: list[Message] | None = None,
        data: dict[str, object] | None = None,
    ) -> list[Message]:
        tasks = [
            agent.work_async(query=query, messages=messages, data=data)
            for agent in self.agents
        ]
        results = await asyncio.gather(*tasks)
        return results


class AgentIndex(BaseModel):
    agents: dict[str, Agent] = Field(default_factory=dict)
    default: Agent | None = None

    def __getitem__(self, key: str) -> Agent:
        return self.agents[key]

    def add(self, name: str, agent: Agent, is_default: bool = False) -> None:
        self.agents[name] = agent
        if is_default:
            self.default = agent

    def remove(self, name: str) -> None:
        if name not in self.agents:
            raise KeyError(f"Agent '{name}' not found in the index.")

        if self.default == self.agents[name]:
            self.default = None

        del self.agents[name]

    def set_default(self, name: str) -> None:
        if name not in self.agents:
            raise KeyError(f"Agent '{name}' not found in the index.")
        self.default = self.agents[name]

    def find(self, names: list[str]) -> list[Agent]:
        return [agent for name in names if (agent := self.agents.get(name)) is not None]


class AgentLegion(BaseModel):
    speaker: Agent
    selector: Agent
    agent_index: AgentIndex

    def _prepare_messages(
        self,
        query: str | None = None,
        messages: list[Message] | None = None,
    ) -> tuple[list[Message], Agent | None]:
        if messages is None:
            messages = []

        selector_result = self.selector.work(query=query, messages=messages)

        assert selector_result.data is not None, "Selector result missing data."
        assert "selections" in selector_result.data, "Selector data missing 'selections' key."

        selections = selector_result.data["selections"]
        selected_agents = self.agent_index.find(selections)

        if len(selected_agents) == 1:
            return messages, selected_agents[0]

        agent_group = AgentGroup(agents=selected_agents)

        results = asyncio.run(
            agent_group.work_async(query=query, messages=messages)
        )

        for agent_name, result in zip(selections, results):
            result.content = (
                f"**{agent_name} agent** response (NOT VISIBLE TO USER):\n\n{result.content}"
            )
            messages.append(result)

        return messages, None

    def work(
        self,
        query: str | None = None,
        messages: list[Message] | None = None,
    ) -> Message:
        messages, single_agent = self._prepare_messages(query=query, messages=messages)

        if single_agent is not None:
            return single_agent.work(query=query, messages=messages)

        return self.speaker.work(messages=messages)

    def stream(
        self,
        query: str | None = None,
        messages: list[Message] | None = None,
    ) -> Iterator[Message]:
        messages, single_agent = self._prepare_messages(query=query, messages=messages)

        if single_agent is not None:
            yield from single_agent.stream(query=query, messages=messages)
            return

        yield from self.speaker.stream(messages=messages)